; ebauche de calcul symbolique
; derivation et simplification d'expressions
; programme ecrit en lisp vers le debut des annees 1980
; adaptations a faire suivant votre lisp ou scheme
; pour evoluer vers un systeme moins fruste
; il faudrait rajouter les rationnels, des entiers (tres) longs
; et utiliser commutativite et distributivite dans la simplification
; htpp://www.pascaland.org/
; http://franck.pissotte.free.fr/

(de ** (x y)
   (cond
    ((= y 1) x)
    ((= y 0) 1)
    ((> y 0)     (* x (** x (- y 1))))
    (t  'mauvais)
   )
)

(de depoloniser (exp)
    (if (atom exp)
        exp
        (list (depoloniser (cadr exp))
              (car exp)
              (depoloniser (caddr exp)))
    )
)

;d'apres langages et systemes JJ Girardot lisp editest p 187
(de poloniser (exp)
  (polo nil (reverse exp)
        '(
         ( < <= = > >= <> )
         ( + - )
         ( * / )
         ( ** )
         )
  )
)

(de polo (x y z)
  (cond
    ((null y)
      (cond
        ((null z)
          (cond
            ((atom (car x)) (car x))
            (t     (poloniser (car x)))
          )
        )
        (t  (polo nil x (cdr z)))
      )
     )
     ((memq (car y) (car z))
       (list (car y)
          (polo nil (cdr y) z)
          (polo nil x       (cdr z))
        )
       )
       (t (polo (append x (cons (car y) nil)) (cdr y) z))
   )
)

;d'apres micro-systemes no 50 p 174
;(de simplifier (EXP)
;    (setq sexp (simplif EXP))
;    (while (nequal EXP sexp)
;           (setq EXP sexp)
;           (setq sexp (simplif exp))
;    )
;    exp
;)

(de simplifier (exp)
   (setq sexp (simplif exp))
   (if (equal exp sexp)
       exp
       (simplifier sexp)
   )
)

(de simpli (op a1 a2)
  (cond
    ((eq op '+ )  (SIMPLIFPLUS a1 A2))
    ((eq op '- )  (SIMPLIFmoin a1 A2))
    ((eq op '* )  (SIMPLIFMULT a1 A2))
    ((eq op '/ )  (SIMPLIFdiv  a1 A2))
    ((eq op '** )  (SIMPLIFPUIS a1 A2))
    (T 'erreur simplif)
  )
)

(de simplif (EXP)
  (if (atom EXP)
      EXP
      (simpli (car EXP) (cadr EXP) (caddr EXP))
  )
)

 (de SIMPLIFPLUS ( A1 A2)
   (cond
     ((equal A1 0) (simplif A2))
     ((equal A2 0) (simplif A1))
     ((numberp A1) (if (numberp A2)
                       (+ A1 A2)
                       (SIMPLIFNUM '+ A1 A2)))
     ((numberp A2) (SIMPLIFNUM '+ A2 A1))
     ((equal A1 A2)(list '* (simplif A1 ) 2))
     (t            (list '+ (simplif A1) (simplif A2)))
))

 (de SIMPLIFmoin ( A1 A2)
   (cond
     ((equal A1 0) (if (numberp a2)
                       (- 0 a2)
                       (list '-
                             0
                             (simplif A2))))
     ((equal A2 0) (simplif A1))
     ((numberp A1) (if (numberp A2)
                       (- A1 A2)
                       (SIMPLIFNUM '- A1 A2)))
     ((numberp A2) (SIMPLIFNUM '- A1 A2))
     ((equal A1 A2) 0)
     (t            (list '- (simplif A1) (simplif A2)))
))

 (de SIMPLIFMULT ( A1 A2)
   (cond
     ((equal A1 0) 0)
     ((equal A2 0) 0)
     ((equal A1 1) (simplif A2))
     ((equal A2 1) (simplif A1))
     ((numberp A1) (if (numberp A2)
                     (* A1 A2)
                     (SIMPLIFNUM '* A1 A2)))
     ((numberp A2) (SIMPLIFNUM '* A2 A1))
     ((equal A1 A2)(list '** (simplif A1 ) 2))
     (t            (list '* (simplif A1) (simplif A2)))
))

 (de SIMPLIFdiv ( A1 A2)
   (cond
     ((equal A1 0) 0)
     ((equal A2 0) 'Division_0)
     ((equal A2 1) (simplif A1))
     ((equal A1 A2) 1)
     ((numberp A1) (if (numberp A2)
                     (/ A1 A2)
                     (SIMPLIFNUM '/ A1 A2)))
     ((numberp A2) (SIMPLIFNUM '/ A1 A2))
     (t            (list '/ (simplif A1) (simplif A2)))
))

(de SIMPLIFPUIS (A1 A2)
  (cond
    ((equal A1 0) 0)
    ((equal A2 0) 1)
    ((equal A2 1) (simplif A1))
    (t            (list '** (simplif A1) (simplif A2)))
))

(de SIMPLIFNUM (OP A1 A2)
  (cond
    ((atom A2)        (list OP A1 A2))
    ((nequal OP (car A2)) (list OP A1 (simplif A2)))
    ((numberp (cadr A2))
                 (list (car A2) (apply OP (list A1 (cadr A2))) (caddr A2)))
    ((numberp (caddr A2))
                 (list (car A2) (apply OP (list A1 (caddr A2))) (cadr A2)))
    (t                     (list OP A1 (simplif A2)))
))

(de deriver (exp var)
    (depoloniser (deriv (poloniser exp) var))
)

(de deriv (EXP VAR)
  (simplifier (deri EXP VAR )))

(de deri (EXP VAR)
(cond
  ( (numberp EXP           ) 0                         )
  ( (atom    EXP           ) (if (equal EXP VAR ) 1 0) )
  ( (equal   (car EXP) '+  ) (list '+ (deri (cadr EXP) VAR)
                                    (deri (caddr EXP) VAR)))
  ( (equal   (car EXP) '-  ) (list '- (deri (cadr EXP) VAR)
                                    (deri (caddr EXP) VAR)))
  ( (equal   (car EXP) '*  ) (list '+
                                 (list '* (cadr EXP)
                                       (deri (caddr EXP) VAR))
                                 (list '* (deri (cadr EXP) VAR)
                                       (caddr EXP))))
  ( (equal   (car EXP) '/  ) (list '+
                                   (list '/
                                         (deri (cadr EXP) var)
                                         (caddr EXP))
                                   (list '*
                                         (cadr EXP)
                                         (deriv_inv (caddr EXP)))
                             ))
  ( (equal   (car EXP) '** ) (DERIVPUIS (cadr EXP) (caddr EXP)))
  ( t                        'operateur_inconnu)
))

(de deriv_inv (exp)
  (list '-
        0
        (list '/
              (deri exp var)
              (list '** exp 2)
        )
) )

(de DERIVPUIS (A1 A2)
   (cond
     ((equal A2 0) 0)
     ((equal A2 1) 0)
     ((equal A2 2) (list '*
                         (list '*
                               (deri A1 VAR )
                               A1
                         )
                         2
                   ))
     (t            (list '*
                         (list '*
                               (deri A1 VAR)
                               (list '** A1  (list '- A2 1))
                         )
                         A2
                   ))
   )
)
